"""Module grouping some useful data-structures and functionality used by other modules
in the `exomole` package.

This module only groups *helper* functions and classes, which are not designed to be
used by the end-users of the `exomole` package.
"""

import warnings
from pathlib import Path

import pandas
import requests

from .exceptions import (
    APIError,
    LineWarning,
    LineCommentError,
    LineValueError,
    DataParseError,
)


def get_file_raw_text_over_api(
    which, molecule_slug=None, isotopologue_slug=None, dataset_name=None
):
    """Get the raw text of any ExoMol file over the ExoMol api.

    Exomol *.def* or *.all* file are supported, controlled by the `which` argument.
    For ``which='all'``, all the other optional arguments are ignored.
    The file is requested over *https* under the relevant URL via the ExoMol public API.

    Parameters
    ----------
    which : {'all', 'def'}
    molecule_slug, isotopologue_slug, dataset_name : str, optional
        Ignored if ``which == 'all'``.

    Returns
    -------
    raw_text : str
        The raw text of the file requested.

    Raises
    ------
    APIError
        If the arguments passed result in a request with an unsuccessful response.
    """
    base_url = "https://www.exomol.com/db/"
    if which == "all":
        url = f"{base_url}/exomol.all"
    elif which == "def" and all([molecule_slug, isotopologue_slug, dataset_name]):
        url = (
            f"{base_url}{molecule_slug}/{isotopologue_slug}/"
            f"{dataset_name}/{isotopologue_slug}__{dataset_name}.def"
        )
    else:
        raise ValueError(f"Unrecognised arguments passed")

    response = requests.get(url)
    if response.status_code != 200:
        raise APIError(f"Unsuccessful response received from {url}")
    raw_text = response.text
    return raw_text


def parse_exomol_line(
    lines,
    n_orig,
    expected_comment=None,
    file_name=None,
    val_type=None,
    warn_on_comments=False,
):
    """A helper line parser for the ExoMol files (*.all* and *.def*).

    List of the file lines is passed as well as the original length of the list
    `n_orig`.
    The top line of `lines` is popped (`lines` gets changed as an externality) and the
    line value is extracted, cast to the final data type (`val_type`) and returned.
    The list of `lines` is therefore being consumed line by line with each call of this
    function.
    If the `expected_comment` is passed, the comment in the top line
    (after the ``#`` symbol) is checked against the `expected_comment`,
    and if they do not match and `warn_on_comment` is set to ``True``, the
    `LineWarning` is raised.
    If the `val_type` is passed, the value is cast to the passed type.

    Parameters
    ----------
    lines : list of str
        At first containing all the lines of the file which are then being
        consumed one by one.
    n_orig : int
        The number of lines of the full file (for error raising only).
    expected_comment : str, optional
        The comment after the ``#`` symbol on each line is expected to match the
        passed `expected_comment`.
    file_name : str, optional
        The name of the file that `lines` belonged to (for error raising only).
    val_type : type, optional
        The intended `type` of the parsed value, the value will be cast to.
    warn_on_comments : bool, optional
        If ``True``, the `LineWarning` will be raised if the parsed comment
        does not match the `expected_comment`.

    Returns
    -------
    str or int or float
        Value belonging to the top line of `lines` passed. Type is either `str`,
        or the `val_type` passed.

    Raises
    ------
    LineCommentError
        If the top line does not have the required format of ``value # comment``
    LineValueError
        If the value parsed from the top line cannot be cast to the `val_type`.

    Warnings
    --------
    LineWarning
        If `warn_on_comment` set to ``True`` and the comment parsed from the top line
        does not match the `expected_comment` passed, the `LineWarning` is raised.
        Also raised if an empty line is detected anywhere (irrespective of the
        `warn_on_comment` value).
    """

    while True:
        try:
            line = lines.pop(0).strip()
        except IndexError:
            msg = f"Run out of lines"
            if file_name:
                msg += f" in {file_name}"
            raise LineValueError(msg)
        line_num = n_orig - len(lines)
        if line:
            break
        else:
            msg = f"Empty line detected on line {line_num}"
            if file_name:
                msg += f" in {file_name}"
            warnings.warn(msg, LineWarning)
    try:
        val, comment = line.split("# ")
        val = val.strip()
    except ValueError:
        msg = f"Unexpected line format detected on line {line_num}"
        if file_name:
            msg += f" in {file_name}"
        raise LineCommentError(msg)
    if val_type:
        try:
            val = val_type(val)
        except ValueError:
            msg = f"Unexpected value type detected on line {line_num}"
            if file_name:
                msg += f" in {file_name}"
            raise LineValueError(msg)
    if expected_comment and warn_on_comments and comment != expected_comment:
        msg = f"Unexpected comment detected on line {line_num}!"
        if file_name:
            msg += f" in {file_name}"
        warnings.warn(msg, LineWarning)
    return val


def load_dataframe_chunks(
    file_path,
    chunk_size,
    first_col_is_index=False,
    column_names=None,
    dtype=None,
    check_num_columns=True,
):
    """Generates chunks of a compressed ExoMol data file.

    Chunks of either *.states.bz2* file or *.trans.bz2* file are loaded from the
    local file system with the specified chunk size.
    Generator of `pandas.DataFrames` is returned.
    No decompression is necessary beforehand.
    If `column_names` are passed, and `check_num_columns` is ``True``, it verifies that
    ``len(column_names)`` matches to the number of columns in the file,
    otherwise an exception is raised.

    Parameters
    ----------
    file_path : str or Path
        Path to the *.bz2* compressed file I want to load. Either *.states* or *.trans*
        file.
    chunk_size : int
        Appropriate value depending on RAM available.
    first_col_is_index : bool, optional
        If ``True``, the first data column values are set as the chunk index (and the
        first column name is therefore ignored - but still must be present in
        `column_names`).
    column_names : list of str, optional
        Column names of the file loaded. If ``first_column_is_index is True``,
        the `column_names` *still have to contain* the index (first) column.
    dtype : type, optional
        Data ``type`` to cast to `pandas.read_csv`. By default, data type determination
        is left to `pandas`.
    check_num_columns : bool, optional
        If ``True`` and `column_names` passed, check is performed to verify that the
        `column_names` are consistent with the number of columns in the data file.
        This check will likely result in some slowdown as the file will be decompressed
        twice.

    Returns
    -------
    df_chunks : pandas.io.parsers.TextFileReader
        Generator of `pandas.DataFrame` chunks. Access by
        ``for chunk in df_chunks: ...``, where each chunk is a `pandas.DataFrame`.

    Raises
    ------
    DataParseError
        When ``check_num_columns is True`` and `column_names` are inconsistent with the
        number of columns in the data file being read.
    """
    if check_num_columns and column_names:
        file_name = Path(file_path).name
        num_cols = get_num_columns(file_path)
        if num_cols != len(column_names):
            raise DataParseError(
                f"{file_name} has {num_cols} columns, but column names "
                f"{column_names} were passed."
            )

    df_chunks = pandas.read_csv(
        file_path,
        compression="bz2",
        sep=r"\s+",
        header=None,
        index_col=None if not first_col_is_index else 0,
        names=column_names
        if not (column_names and first_col_is_index)
        else column_names[1:],
        chunksize=chunk_size,
        iterator=True,
        low_memory=False,
        dtype=dtype,
    )
    return df_chunks


def get_num_columns(file_path):
    """Gets the number of columns in the *.bz2* compressed either *.states*, or
    *.trans* file under the `file_path`.

    Parameters
    ----------
    file_path : str or Path

    Returns
    -------
    int
    """
    for chunk in load_dataframe_chunks(file_path, chunk_size=1):
        _, num_cols = chunk.shape
        return int(num_cols)


class DataClass:
    """Base class for all the data-classes used to store data from the parsed *.all*
    and *.def* files."""

    def __init__(self, **kwargs):
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def __repr__(self):
        cls_name = self.__class__.__name__
        attrs_str = ", ".join(f"{attr}={val}" for attr, val in vars(self).items())
        return f"{cls_name}({attrs_str})"



"""Module grouping some data-classes and a parser for reading and parsing the
ExoMol master file *exomol.all*.
"""

import warnings
from pathlib import Path

from pyvalem.formula import Formula, FormulaParseError

from .exceptions import AllParseError, AllParseWarning, LineValueError, LineCommentError
from .utils import DataClass
from .utils import get_file_raw_text_over_api, parse_exomol_line


# noinspection PyUnresolvedReferences
class Molecule(DataClass):
    """A data class representing the molecule instance.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    names : list of str
    formula : str
    isotopologues : dict of Isotopologue
        The `Isotopologue` instances are stored under the keys of
        `Isotopologue.formula`.
    """

    def __init__(self, names, formula, isotopologues):
        super().__init__(names=names, formula=formula, isotopologues=isotopologues)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.formula})"


# noinspection PyUnresolvedReferences
class Isotopologue(DataClass):
    """A data class representing a molecule instance.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    inchi_kay : str
    iso_slug : str
    iso_formula : str
    dataset_name : str
    version : int
    """

    def __init__(self, inchi_key, iso_slug, iso_formula, dataset_name, version):
        super().__init__(
            inchi_key=inchi_key,
            iso_slug=iso_slug,
            iso_formula=iso_formula,
            dataset_name=dataset_name,
            version=version,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.iso_slug})"


class AllParser:
    """Class which handles parsing the *exomol.all* file.

    Parses the *.all* file specified by the `path` argument passed and leading to
    the *.all* file on the local file system. If the `path` is not given, the *.all*
    file is requested via the ExoMol public API.
    Instantiating the class only saves the `raw_text` attribute, which gets parsed
    with the `parse` method into all the available data structured.
    All the *relevant* attributes are listed in the **Attributes** section.

    Parameters
    ----------
    path : str or Path, optional
        Path to the *exomol.all* file. If not passed, the file is requested over
        the ExoMol public API.

    Attributes
    ----------
    raw_text : str
    file_name : str
    version : int
    molecules : dict of Molecule
        The `Molecule` instances are stored under the keys of `Molecule.formula`.

    Raises
    ------
    APIError
        If `path` not passed and the ExoMol API request call results in an unsuccessful
        response.

    Notes
    -----
    See the ExoMol file standard as defined in the ExoMol release paper [1]_.

    References
    ----------
    .. [1] Tennyson J, et al. The ExoMol database: molecular line lists for
       exoplanet and other hot atmospheres. J Mol Spectrosc 2016;327:73–94.
       doi: 10.1016/j.jms.2016.05.002

    Examples
    --------
    Instantiate the parser:
    >>> parser = AllParser(path="tests/resources/exomol_data/exomol.all")
    >>> parser.file_name
    'exomol.all'
    >>> parser.raw_text[: 13]  # first 13 characters of the text
    'EXOMOL.master'

    Parse the exomol.all text:
    >>> parser.parse(warn_on_comments=True)
    >>> parser.id
    'EXOMOL.master'
    >>> parser.version
    20210707
    >>> molecule_parsed = parser.molecules['CaH']
    >>> type(molecule_parsed)
    <class 'exomole.read_all.Molecule'>
    >>> molecule_parsed.names
    ['Calcium monohydride', 'Calcium(I) hydride']
    >>> isotopologue_parsed = molecule_parsed.isotopologues['(40Ca)(1H)']
    >>> isotopologue_parsed
    Isotopologue(40Ca-1H)
    >>> type(isotopologue_parsed)
    <class 'exomole.read_all.Isotopologue'>
    >>> isotopologue_parsed.dataset_name
    'Yadin'
    >>> isotopologue_parsed.iso_slug
    '40Ca-1H'
    """

    def __init__(self, path=None):
        self.raw_text = None
        self.file_name = None
        self._save_raw_text(path)
        # placeholders for all the attributes
        self.id = None
        self.version = None
        self.num_molecules = None
        self.num_isotopologues = None
        self.num_datasets = None
        self.molecules = None

    def _save_raw_text(self, path):
        """Save the raw text of a *.all* file as an instance attribute.

        The *.all* file is either read from the local file system, or requested over the
        ExoMol public API, if `path` argument not passed.

        Parameters
        ----------
        path : str or Path, optional
            Path leading to the *.all* file. If not supplied, the file is requested over
            the ExoMol public API.
        """
        if path is None:
            self.raw_text = get_file_raw_text_over_api("all")
            self.file_name = "exomol.all"
        else:
            with open(path, "r") as fp:
                self.raw_text = fp.read()
            self.file_name = Path(path).name

    def parse(self, warn_on_comments):
        """Parse the *.all* file text from the `raw_text` attribute.

        Populates all the instance attributes incrementally, util it hits the end of
        the file, or one of the exceptions is raised, signaling inconsistent *.all*
        file.

        Parameters
        ----------
        warn_on_comments : bool
            If ``True``, the comments behind the ``#`` symbol on each line are checked
            against some expected comments (hard-coded in the method) and the
            `LineWarning` is raised if they do not match.

        Raises
        -------
        AllParseError
            Raised if value on any line cannot be cast to the expected `type`, or if
            the parser runs out of lines. This error signals an inconsistent *.all*
            file. Also raised when any other inconsistencies are detected, such as
            formulas not supported by the `PyValem` package, etc.

        Warns
        -----
        AllParseWarning
            This warning is raised if `num_isotopologues` or `num_datasets`
            do not agree with the the actual numbers of isotopologues and datasets
            respectively, extracted from the file lines. Also raised if the *.all*
            file lists more than a single dataset for any given isotopologue.
        LineWarning
            Raised if `warns_on_comments` is ``True`` and if the comment on any line
            does not match the expected text hard-coded in this method.
        """
        lines = self.raw_text.split("\n")
        n_orig = len(lines)

        def parse_line(expected_comment, val_type=None):
            return parse_exomol_line(
                lines,
                n_orig,
                expected_comment=expected_comment,
                file_name=self.file_name,
                val_type=val_type,
                warn_on_comments=warn_on_comments,
            )

        # catch all the parse_line-originated errors and wrap them in the AllParseError:
        try:
            self.id = parse_line("ID")
            if self.id != "EXOMOL.master":
                raise AllParseError(f"Unexpected ID in {self.file_name}")
            self.version = parse_line("Version number with format YYYYMMDD", int)
            self.num_molecules = parse_line("Number of molecules in the database", int)
            self.num_isotopologues = parse_line(
                "Number of isotopologues in the database", int
            )
            self.num_datasets = parse_line("Number of datasets in the database", int)
            self.molecules = {}

            # Verify the numbers of isotopologues and datasets by keeping track:
            all_isotopologues = []
            all_datasets = []
            # Also keep track of all the molecules with more than one dataset in a
            # single isotopologue:
            molecules_with_duplicate_isotopologues = []

            # loop over molecules:
            for _ in range(self.num_molecules):
                mol_names = []

                num_names = parse_line("Number of molecule names listed", int)

                # loop over the molecule names:
                for __ in range(num_names):
                    mol_names.append(parse_line("Name of the molecule"))

                mol_formula = parse_line("Molecule chemical formula")
                try:
                    Formula(mol_formula)
                except FormulaParseError as e:
                    raise AllParseError(f"{str(e)} (raised in {self.file_name})")

                num_isotopologues = parse_line(
                    "Number of isotopologues considered", int
                )
                mol_isotopologues = {}

                # loop over the isotopologues:
                for __ in range(num_isotopologues):
                    iso_inchi_key = parse_line("Inchi key of isotopologue")
                    iso_slug = parse_line("Iso-slug")
                    iso_formula = parse_line("IsoFormula")
                    try:
                        Formula(iso_formula)
                    except FormulaParseError as e:
                        raise AllParseError(f"{str(e)} (raised in {self.file_name})")
                    iso_dataset_name = parse_line("Isotopologue dataset name")
                    iso_version = parse_line("Version number with format YYYYMMDD", int)

                    isotopologue = Isotopologue(
                        inchi_key=iso_inchi_key,
                        iso_slug=iso_slug,
                        iso_formula=iso_formula,
                        dataset_name=iso_dataset_name,
                        version=iso_version,
                    )

                    if iso_formula not in mol_isotopologues:
                        mol_isotopologues[iso_formula] = isotopologue
                    else:
                        warnings.warn(
                            f"{mol_formula} lists more than one dataset for "
                            f"isotopologue {iso_formula}. Ignoring {iso_dataset_name}",
                            AllParseWarning,
                        )
                        molecules_with_duplicate_isotopologues.append(mol_formula)

                    all_datasets.append(iso_dataset_name)
                    all_isotopologues.append(isotopologue)

                # molecule slug is not present in the exomol.all data!
                self.molecules[mol_formula] = Molecule(
                    names=mol_names,
                    formula=mol_formula,
                    isotopologues=mol_isotopologues,
                )
        except (LineValueError, LineCommentError) as e:
            raise AllParseError(str(e))

        if self.num_isotopologues != len(all_isotopologues):
            warnings.warn(
                f"Number of isotopologues stated ({self.num_isotopologues}) does not "
                f"match the actual number ({len(all_isotopologues)})!",
                AllParseWarning,
            )

        if self.num_datasets != len(set(all_datasets)):
            warnings.warn(
                f"Number of datasets stated ({self.num_datasets}) does not match the "
                f"actual number ({len(set(all_datasets))})!",
                AllParseWarning,
            )



"""
Module containing all the custom Exceptions and Warnings used in the `exomole` package.
"""


class APIError(Exception):
    pass


class LineCommentError(Exception):
    pass


class LineValueError(Exception):
    pass


class LineWarning(UserWarning):
    pass


class AllParseError(Exception):
    pass


class AllParseWarning(UserWarning):
    pass


class DefParseError(Exception):
    pass


class DataParseError(Exception):
    pass


class StatesParseError(DataParseError):
    pass


class TransParseError(DataParseError):
    pass



"""Module grouping some data-classes and the parser for reading and parsing the
ExoMol *.def* files.
"""
import warnings
from pathlib import Path

from pyvalem.formula import Formula, FormulaParseError

from .exceptions import LineValueError, LineCommentError, LineWarning, DefParseError
from .utils import DataClass
from .utils import get_file_raw_text_over_api, parse_exomol_line


# noinspection PyUnresolvedReferences
class Isotope(DataClass):
    """A data class representing isotope instances.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    number : int
    element_symbol : str
    """

    def __init__(self, number, element_symbol):
        super().__init__(number=number, element_symbol=element_symbol)

    def __repr__(self):
        return f"Isotope({self.number}{self.element_symbol})"


class IrreducibleRepresentation(DataClass):
    """A data class representing instances of irreducible representations.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    ir_id : str
        ID of the *irreducible representation*
    label : str
    nuclear_spin_degeneracy : int
    """

    def __init__(self, ir_id, label, nuclear_spin_degeneracy):
        super().__init__(
            id=ir_id, label=label, nuclear_spin_degeneracy=nuclear_spin_degeneracy
        )


class QuantumCase(DataClass):
    """A data class representing the quantum case instances.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    label : str
    """

    def __init__(self, label):
        super().__init__(label=label)


# noinspection PyUnresolvedReferences
class Quantum(DataClass):
    """A data class representing the quanta instances.

    All the parameters passed are stored as instance attributes.

    Parameters
    ----------
    label : str
    q_format : str
        The quantum format as specified by the *.def* file
    description : str
    """

    def __init__(self, label, q_format, description):
        super().__init__(label=label, format=q_format, description=description)

    def __repr__(self):
        return f"Quantum({self.label})"


class DefParser:
    """Class handling parsing of any particular *.def* file.

    Parses the *.def* file specified either by the `path` argument passed and leading to
    the *.def* file on the local file system, or by the trio of `molecule_slug`,
    `isotopologue_slug` and `dataset_name` arguments, in which case the *.def* file
    is requested via the ExoMol public API.
    Instantiating the class only saves the `raw_text` attribute, which can be parsed
    with the `parse` method into all the available info. All the *relevant* attributes
    are listed in the **Attributes** section.

    Parameters
    ----------
    path : str or Path, optional
        The path leading to the *.def* file. If supplied, all the other arguments are
        simply ignored.
    molecule_slug : str, optional
        Only required, if the `path` argument is not passed.
    isotopologue_slug : str, optional
        Only required, if the `path` argument is not passed.
    dataset_name : str, optional
        Only required, if the `path` argument is not passed.

    Attributes
    ----------
    raw_text : str
    file_name : str
    version : int
    iso_formula : str
    iso_slug : str
    isotopes : list of Isotope
    lifetime_availability : bool
    lande_factor_availability : bool
    quanta : list of Quantum

    Raises
    ------
    APIError
        If `path` not passed and the ExoMol API request call results in an unsuccessful
        response.

    Notes
    -----
    See the ExoMol file standard as defined in the **ExoMol release paper** [1]_.

    References
    ----------
    .. [1] Tennyson J, et al. The ExoMol database: molecular line lists for
       exoplanet and other hot atmospheres. J Mol Spectrosc 2016;327:73–94.
       doi: 10.1016/j.jms.2016.05.002

    Examples
    --------
    Instantiate the parser:
    >>> parser = DefParser(
    ...     path="tests/resources/exomol_data/CaH/40Ca-1H/Yadin/40Ca-1H__Yadin.def"
    ... )
    >>> parser.file_name
    '40Ca-1H__Yadin.def'
    >>> parser.raw_text[:10]  # first 10 characters of the text
    'EXOMOL.def'

    Parse the .def file:
    >>> parser.parse(warn_on_comments=True)
    >>> parser.id
    'EXOMOL.def'
    >>> parser.iso_formula
    '(40Ca)(1H)'
    >>> parser.mass
    40.970416
    >>> parser.isotopes
    [Isotope(40Ca), Isotope(1H)]
    >>> quanta = parser.quanta
    >>> quanta
    [Quantum(par), Quantum(v), Quantum(N), Quantum(e/f)]
    >>> quanta[0].description
    "total parity: '+' or '-'"
    >>> parser.lifetime_availability, parser.lande_factor_availability
    (True, False)

    Additional methods on the parsed data:
    >>> parser.get_quanta_labels()
    ['par', 'v', 'N', 'e/f']
    >>> # with parser.lifetime_availability, we expect 9 columns in the .states file
    >>> parser.get_states_header()
    ['i', 'E', 'g_tot', 'J', 'tau', 'par', 'v', 'N', 'e/f']
    """

    def __init__(
        self,
        path=None,
        molecule_slug=None,
        isotopologue_slug=None,
        dataset_name=None,
    ):
        self.raw_text = None
        self.file_name = None
        self._save_raw_text(path, molecule_slug, isotopologue_slug, dataset_name)
        # placeholders to all the attributes
        self.id = None
        self.iso_formula = None
        self.iso_slug = None
        self.dataset_name = None
        self.version = None
        self.inchi_key = None
        self.isotopes = None
        self.mass = None
        self.symmetry_group = None
        self.irreducible_representations = None
        self.max_temp = None
        self.num_pressure_broadeners = None
        self.dipole_availability = None
        self.num_cross_sections = None
        self.num_k_coefficients = None
        self.lifetime_availability = None
        self.lande_factor_availability = None
        self.num_states = None
        self.quanta_cases = None
        self.quanta = None
        self.num_transitions = None
        self.num_trans_files = None
        self.max_wavenumber = None
        self.high_energy_complete = None

    def _save_raw_text(self, path, molecule_slug, isotopologue_slug, dataset_name):
        """Save the raw text of a *.def* file as an instance attribute

        The *.def* file is either read from the local file system, or requested over the
        ExoMol public API, based on the attributes values.

        Parameters
        ----------
        path : str or Path, optional
            Path leading to the *.def* file. If supplied, all the other arguments are
            ignored.
        molecule_slug : str, optional
            Ignored if `path` is passed.
        isotopologue_slug : str or None
            Ignored if `path` is passed.
        dataset_name : str, optional
            Ignored if `path` is passed.
        """
        if path is None:
            self.raw_text = get_file_raw_text_over_api(
                "def", molecule_slug, isotopologue_slug, dataset_name
            )
            self.file_name = f"{isotopologue_slug}__{dataset_name}.def"
        else:
            with open(path, "r") as fp:
                self.raw_text = fp.read()
            self.file_name = Path(path).name

    def parse(self, warn_on_comments):
        """Parse the *.def* file text from the `raw_text` attribute.

        Populates all the instance attributes incrementally, util it hits the end of
        the file, or one of the exceptions is raised, signaling inconsistent *.def*
        file.

        Parameters
        ----------
        warn_on_comments : bool
            If ``True``, the comments behind the ``#`` symbol on each line are checked
            against some expected comments (hard-coded in the method) and the
            `LineWarning` is raised if they do not match.

        Raises
        -------
        DefParseError
            Raised if value on any line cannot be cast to the expected ``type``, or if
            the parser runs out of lines. This error signals an inconsistent *.def*
            file. Also raised when any other inconsistencies are detected, such as
            formula not supported by the `PyValem` package, etc.

        Warns
        -----
        LineWarning
            Raised if `warns_on_comments` is ``True`` and if the comment on any line
            does not match the expected text hard-coded in this method.
            Also raised if any empty line is present in the .def file, or of the
            number of isotope sections does not match the number of atoms in the
            isotopologue.

        Warnings
        --------
        Currently the parser stops after the *High Energy Complete* line and does not
        parse the rest of the *.def* file, as the info beyond this point in the *.def*
        file was not needed for the data product application which served as my
        motivation to write this package.
        """
        lines = self.raw_text.split("\n")
        n_orig = len(lines)

        def parse_line(expected_comment, val_type=None, local_lines=lines):
            return parse_exomol_line(
                local_lines,
                n_orig,
                expected_comment=expected_comment,
                file_name=self.file_name,
                val_type=val_type,
                warn_on_comments=warn_on_comments,
            )

        # catch all the parse_line-originated errors and wrap them in a higher-level
        # error:
        try:
            self.id = parse_line("ID")
            if self.id != "EXOMOL.def":
                raise DefParseError(f"Unexpected ID in {self.file_name}")
            self.iso_formula = parse_line("IsoFormula")
            self.iso_slug = parse_line("Iso-slug")
            self.dataset_name = parse_line("Isotopologue dataset name")
            self.version = parse_line("Version number with format YYYYMMDD", int)
            self.inchi_key = parse_line("Inchi key of molecule")
            self.isotopes = []
            num_atoms = parse_line("Number of atoms", int)
            try:
                formula = Formula(self.iso_formula)
            except FormulaParseError as e:
                raise DefParseError(f"{str(e)} (raised in {self.file_name})")
            if formula.natoms != num_atoms:
                raise DefParseError(f"Incorrect number of atoms in {self.file_name}")
            # many (probably all) .def files for polyatomic datasets actually do not
            # list all isotopes, but rather only all *distinct* isotopes.from
            # I'll handle this with a Warning.
            num_distinct_atoms = len(formula.atoms)

            def add_isotope(num, el_symbol):
                try:
                    Formula(f"({num}{el_symbol})")
                except FormulaParseError as exc:
                    raise DefParseError(f"{str(exc)} (raised in {self.file_name})")
                isotope = Isotope(number=num, element_symbol=el_symbol)
                self.isotopes.append(isotope)

            i = 0
            for i in range(num_distinct_atoms):
                number = parse_line(f"Isotope number {i + 1}", int)
                element_symbol = parse_line(f"Element symbol {i + 1}")
                add_isotope(number, element_symbol)
            if num_distinct_atoms < num_atoms:
                num_additional_isotopes_expected = num_atoms - num_distinct_atoms
                lines_clone = lines.copy()
                try:
                    for j in range(i + 1, i + 1 + num_additional_isotopes_expected):
                        number = parse_line(
                            f"Isotope number {j + 1}", int, local_lines=lines_clone
                        )
                        element_symbol = parse_line(
                            f"Element symbol {j + 1}", local_lines=lines_clone
                        )
                        add_isotope(number, element_symbol)
                except (LineValueError, LineCommentError):
                    # This means that the .def file lists only distinct isotopes, not
                    # all isotopes, as it should. Handle with Warning and continue with
                    # the original lines
                    warnings.warn(
                        f"Incorrect number of isotopes listed in {self.file_name}",
                        LineWarning,
                    )
                else:
                    # This means that the try clause did not raise anything, meaning
                    # all the isotopes were listed in the .def file. Need to sync
                    # lines with the lines_clone:
                    for _ in range(2 * num_additional_isotopes_expected):
                        lines.pop(0)

            iso_mass_amu = float(
                parse_line("Isotopologue mass (Da) and (kg)").split()[0]
            )
            self.mass = iso_mass_amu
            self.symmetry_group = parse_line("Symmetry group")
            self.irreducible_representations = []
            num_irreducible_representations = parse_line(
                "Number of irreducible representations", int
            )
            for _ in range(num_irreducible_representations):
                ir_id = parse_line("Irreducible representation ID", int)
                label = parse_line("Irreducible representation label")
                nuclear_spin_degeneracy = parse_line("Nuclear spin degeneracy", int)
                ir = IrreducibleRepresentation(
                    ir_id=ir_id,
                    label=label,
                    nuclear_spin_degeneracy=nuclear_spin_degeneracy,
                )
                self.irreducible_representations.append(ir)
            self.max_temp = parse_line("Maximum temperature of linelist", float)
            self.num_pressure_broadeners = parse_line(
                "No. of pressure broadeners available", int
            )
            self.dipole_availability = bool(
                parse_line("Dipole availability (1=yes, 0=no)", int)
            )
            self.num_cross_sections = parse_line(
                "No. of cross section files available", int
            )
            self.num_k_coefficients = parse_line(
                "No. of k-coefficient files available", int
            )
            self.lifetime_availability = bool(
                parse_line("Lifetime availability (1=yes, 0=no)", int)
            )
            self.lande_factor_availability = bool(
                parse_line("Lande g-factor availability (1=yes, 0=no)", int)
            )
            self.num_states = parse_line("No. of states in .states file", int)
            self.quanta_cases = []
            self.quanta = []
            num_quanta_cases = parse_line("No. of quanta cases", int)
            # TODO: It is not entirely clear if num_quanta and related blocks should
            #       be nested under a quanta case, or not.
            #       If they are, the data structures need to be changed, and the
            #       parser tweaked.
            for _ in range(num_quanta_cases):
                self.quanta_cases.append(
                    QuantumCase(label=parse_line("Quantum case label"))
                )
            num_quanta = parse_line("No. of quanta defined", int)
            for i in range(num_quanta):
                label = parse_line(f"Quantum label {i + 1}")
                q_format = parse_line(f"Format quantum label {i + 1}")
                description = parse_line(f"Description quantum label {i + 1}")
                quantum = Quantum(
                    label=label, q_format=q_format, description=description
                )
                self.quanta.append(quantum)
            self.num_transitions = parse_line("Total number of transitions", int)
            self.num_trans_files = parse_line("No. of transition files", int)
            self.max_wavenumber = parse_line("Maximum wavenumber (in cm-1)", float)
            self.high_energy_complete = parse_line(
                "Higher energy with complete set of transitions (in cm-1)", float
            )
            # This is where I finish now, it would be nice to sort out the rest of the
            # file one day!
        except (LineValueError, LineCommentError) as e:
            raise DefParseError(str(e))

    def get_quanta_labels(self):
        """Quanta labels for all the quanta extracted from the parsed *.def* file.

        The `parse` method must have been called first and finished without errors.

        Returns
        -------
        list of str
        """
        return [q.label for q in self.quanta]

    def get_states_header(self):
        """Get the header (column names) for the associated *.states* file

        The column names consist of the mandatory part, canonically named as
        ["i", "E", "g_tot", "J"], followed by optional columns "tau", "g_J" and by
        the quanta labels.

        Returns
        -------
        list of str
        """
        states_header = ["i", "E", "g_tot", "J"]
        if self.lifetime_availability:
            states_header.append("tau")
        if self.lande_factor_availability:
            states_header.append("g_J")
        states_header.extend(self.get_quanta_labels())
        return states_header



"""Module containing functionality for reading ExoMole data files.

Two stand-alone functions are provided for reading *bz2*-compressed ExoMol files
with states (*.states.bz2* files) and with transitions (*.trans.bz2*).
"""

from pathlib import Path

from .exceptions import DataParseError, StatesParseError, TransParseError
from .utils import load_dataframe_chunks, get_num_columns


def states_chunks(states_path, chunk_size, columns):
    """
    Get a generator of chunks of the dataset *.states.bz2* file.

    Generator of `pandas.DataFrame` chunks of the *.states* file, with
    rows indexed by the values of the first column in the *.states* file.

    The `columns` argument passed needs to contain names for *all* the columns
    *including the first* column, which is assumed to be the `states` index.
    The first column `columns[0]` needs to be "i" as so it is explicitly asserted
    that it is included.

    The generated `pandas.DataFrames` are cast explicitly to ``dtype=str``,
    to avoid possible nasty surprises caused by `pandas` guessing the types itself.
    The index, however, is cast to "int64" explicitly.
    The columns can be re-casted downstream to the more appropriate data types
    for faster processing. An example for the energy column might be as follows:
    ``state_chunk['E'] = state_chunk['col'].astype('float64')``

    Parameters
    ----------
    states_path : str or Path
        Path to the *.states* file on the local file system.
    chunk_size : int
        Chunk size, should be chosen appropriately with regards to the RAM size.
        Roughly 1_000_000 per 1GB consumed.
    columns : iterable of str
        Column names for all the columns in the *.states* file including the (first)
        index column named "i".
        Therefore, ``len(columns)`` must be equal the number of actual columns
        in the *.states* file.

    Yields
    ------
    states_chunk : pandas.DataFrame
        Generated chunks of the *.states* file, each is a `pandas.DataFrame` with
        columns according to the `columns` passed, and indexed by the values in the
        first column in the *.states* file.
        The whole `DataFrame` is of string (``"O"``) data type, except the index, which
        is ``"int64"``.

    Raises
    ------
    StatesParseError
        If ``len(columns)`` inconsistent with the number of columns in the *.states*
        file.

    Examples
    --------
    >>> sp = "tests/resources/dummy_states_10x5_int_float_int_str_int.states.bz2"
    >>> states_columns = ["i", "col1", "col2", "col3", "col4"]
    >>> for df in states_chunks(states_path=sp, chunk_size=5, columns=states_columns):
    ...     print(df)  # synthetic, unphysical data as an example
    ...     break
                      col1 col2 col3 col4
    1   0.4745999608668017   88    a    4
    2  0.47729879282298115   90    b    7
    3  0.32392489118966217   57    c    6
    4   0.4704792592345922   95    d    1
    5   0.8168636898850669    6    e    9
    """
    if columns[0] != "i":
        raise StatesParseError("The first column of any .states file needs to be 'i'.")
    try:
        chunks = load_dataframe_chunks(
            file_path=states_path,
            chunk_size=chunk_size,
            first_col_is_index=True,
            column_names=columns,
            dtype=str,
            check_num_columns=True,
        )
    except DataParseError as e:
        raise StatesParseError(str(e))
    for chunk in chunks:
        chunk.index = chunk.index.astype("int64")
        yield chunk


def trans_chunks(trans_paths, chunk_size):
    """
    Get a generator of chunks of the dataset *.trans.bz* files.

    Generator of `pandas.DataFrame` chunks of all the *.trans* files passed as the
    `trans_paths` argument value.
    The columns are auto-named as ``"i", "f", "A_if" [, "v_if"]``.
    The ``"i"`` and ``"f"`` columns will correspond to the index of the `DataFrames`
    yielded by the `states_chunks` generator.
    No explicit data type casting is performed and `pandas` is trusted to correctly
    identify the ``"i"`` and ``"f"`` columns as ``"int64"`` and rest as ``"float64"``
    data types.

    Parameters
    ----------
    trans_paths : iterable of (str or Path)
        Paths to the *.trans* files on the local file system. They all need to belong to
        the same dataset, but no checks are made to assert that!
    chunk_size : int
        Chunk size, should be chosen appropriately with regards to RAM size, roughly
        10_000_000 per 1GB consumed.

    Yields
    ------
    trans_chunk : pd.DataFrame
        Generated chunks of all the *.trans* files, each is a `pandas.DataFrame` with
        auto-named columns.

    Raises
    ------
    TransParseError
        If the first *.trans* file has number of columns other than ``{3, 4}``.

    Examples
    --------
    >>> tr_paths = sorted(Path("tests/resources").glob("*.trans*0*.bz2"))
    >>> for tr_path in tr_paths:
    ...     print(tr_path)
    tests/resources/dummy_trans_5x4_int_int_float_float.trans01.bz2
    tests/resources/dummy_trans_5x4_int_int_float_float.trans02.bz2
    tests/resources/dummy_trans_5x4_int_int_float_float.trans03.bz2

    >>> for df in trans_chunks(trans_paths=tr_paths, chunk_size=3):
    ...     print(df)  # synthetic, unphysical data as an example
    ...     break
       i  f      A_if      v_if
    0  7  5  0.275477  0.121660
    1  6  8  0.446633  0.290420
    2  2  8  0.723996  0.426885
    """
    trans_paths = sorted(trans_paths)

    num_cols = get_num_columns(trans_paths[0])
    columns = ["i", "f", "A_if"]
    if num_cols == 4:
        columns.append("v_if")
    elif num_cols != 3:
        raise TransParseError(
            f"Unexpected number of columns in {Path(trans_paths[0]).name}: {num_cols}"
        )
    assert num_cols in {3, 4}
    # yield all the chunks from all the files:
    for file_path in trans_paths:
        chunks = load_dataframe_chunks(
            file_path=file_path, chunk_size=chunk_size, column_names=columns
        )
        for chunk in chunks:
            yield chunk



from setuptools import setup, find_packages
from pathlib import Path

root = Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (root / "README.rst").read_text(encoding="utf-8")

setup(
    name="exomole",
    version="1.2.3",
    description="A package for parsing and validation of Exomol Database data files",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/hanicinecm/exomole/",
    author="Martin Hanicinec",
    author_email="hanicinecm@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    keywords="exomol",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=["pandas", "requests", "pyvalem>=2.3"],
    extras_require={
        "dev": ["pytest-cov", "tox", "black", "ipython"],
    },
    project_urls={
        "Bug Reports": "https://github.com/hanicinecm/exomole/issues",
    },
)
